"""
(c) Copyright 2015 Hexoskin
Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted,
provided that the above copyright notice and this permission notice appear in all copies. The software is provided
"as is" and the author disclaims all warranties with regard to this software including all implied warranties of
merchantability and fitness. In no event shall the author be liable for any special, direct, indirect, or
consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether in an action of
contract, negligence or other tortious action, arising out of or in connection with the use or performance of this
software.

"""

import os
import wave
import json
import copy
import csv
import struct
VERSION = [0, 10, 3]

DEFAULT_SETTINGS = dict(
    add_quality=False,
    group=False,
    group_raw=False,
    epoch=False,
    header=True,
    skip_raw=False,
    as_nan=False,
    delimiter_tab=False
)

one_hz = ['heart_rate', 'heart_rate_quality', 'breathing_rate', 'breathing_rate_quality', 'minute_ventilation_cl',
          'tidal_volume_ml', 'activity', 'cadence', 'temperature_celcius', 'systolic_pressure', 'SPO2', 'SPO2_quality',
          'BATT_ST_CHANNEL_CHAR', 'TEMP_ST_CHANNEL_CHAR', 'energy_mifflin_keytel']
raw = ['ECG_I', 'ECG_II', 'ECG_III', 'respiration_thoracic', 'respiration_abdominal', 'acceleration_X',
       'acceleration_Y', 'acceleration_Z', 'PPG', 'ppgdo', 'ppgDisc' 'ppgOOT']

keys_list = ['filename_in', 'filename_out', 'quality_signal_name', 'unit', 'dt', 'dt_offset',
             'gain', 'offset', 'signed', 'round', 'datatype', 'is_wav']
_FREQ = None


def get_freq():
    return _FREQ


def set_freq(freq):
    global _FREQ
    if freq not in [256, 1000]:
        raise ValueError('Invalid freq')
    _FREQ = freq


def files_specs_table(freq):
    GAIN_ECG = 0.0064 if freq == 256 else 0.001
    OFFSET_ECG = -1360 * 0.0064 if freq == 256 else 0
    specs = [
        # filename_in, csv_name, quality_signal_name, unit, dt,dt_offset, gain, offset, signed, round, datatype, is_wav
        ("ECG_I", "ECG_I", "", "mV", 'NA', 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4113, True),  # ECG_I,
        ("ECG_II", "ECG_II", "", "mV", 'NA', 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4114, True),  # ECG_II,
        ("ECG_III", "ECG_III", "", "mV", 'NA', 0, GAIN_ECG, OFFSET_ECG, 1, 5, 4115, True),  # ECG_III,
        ("respiration_thoracic", "respiration_thoracic", "", "na", 2. / 256, 0, 1, 0, 0, 5, 4129, True),
        ("respiration_abdominal", "respiration_abdominal", "", "na", 2. / 256, 0, 1, 0, 0, 5, 4130, True),
        ("acceleration_X", "acceleration_X", "", "G", 4. / 256, 0, 1 / 256., 0, 1, 8, 4145, True),  # acceleration_X,
        ("acceleration_Y", "acceleration_Y", "", "G", 4. / 256, 0, 1 / 256., 0, 1, 8, 4146, True),  # acceleration_Y,
        ("acceleration_Z", "acceleration_Z", "", "G", 4. / 256, 0, 1 / 256., 0, 1, 8, 4147, True),  # acceleration_Z,
        ("ppgdo", "ppgdo", "", "na", 4. / freq, 0, 1, 0, 0, 5, 64, True),  # ppgdo,
        ("ppgDisc", "ppgDisc", "", "na", 4. / freq, 0, 1, 0, 0, 0, 64, True),  # ppgDisc, DEPRECATED
        ("ppgOOT", "ppgOOT", "", "na", 4. / freq, 0, 1, 0, 0, 5, 64, True),  # ppgOOT, DEPRECATED
        ("heart_rate", "heart_rate", "heart_rate_quality", "bpm", 1., 1, 1, 0, 0, 0, 19, True),  # heart_rate,
        ("energy_mifflin_keytel", "energy_mifflin_keytel", "", "watt", 1., 1, 1, 0, 0, 0, 42, True),
        ("heart_rate_quality", "heart_rate_quality", "NONE", "na", 1., 1, 1, 0, 0, 0, 1000, True),
        # heart_rate_quality,
        ("breathing_rate", "breathing_rate", "breathing_rate_quality", "rpm", 1., 1, 1, 0, 0, 0, 33, True),
        ("breathing_rate_quality", "breathing_rate_quality", "NONE", "na", 1., 1, 1, 0, 0, 0, 1001, True),
        ("minute_ventilation", "minute_ventilation_raw", "", "ml/min", 1., 1, 13.28, 0, 0, 5, 36, True),
        ("tidal_volume", "tidal_volume_raw", "", "ml", 1., 1, 13.28, 0, 0, 5, 37, True),  # tidal_volume, # DEPRECATED
        ("minute_ventilation_adjusted", "minute_ventilation_adjusted", "", "ml/min", 1., 1, 10, 0, 0, 5, 38, True),
        ("tidal_volume_adjusted", "tidal_volume_adjusted", "", "ml", 1., 1, 1, 0, 0, 5, 39, True),
        ("minute_ventilation_cl", "minute_ventilation_adjusted", "", "ml/min", 1., 1, 10, 0, 0, 5, 38, True),
        ("tidal_volume_ml", "tidal_volume_adjusted", "", "ml", 1., 1, 1, 0, 0, 5, 39, True),  # tidal_volume_ml,
        ("activity", "activity", "", "G", 1., 1, 1 / 256., 0, 1, 8, 49, True),  # activity,
        ("cadence", "cadence", "", "spm", 1., 1, 1, 0, 0, 0, 53, True),  # cadence,
        ("temperature_celcius", "temperature_celcius", "", "Celcius", 1., 1, 1 / 256., 0, 1, 8, 81, True),
        ("systolic_pressure", "systolic_pressure", "", "mmHg", 1., 1, 1, 0, 0, 5, 98, True),  # systolic_pressure
        ("SPO2", "SPO2", "SPO2_quality", "Percent", 1., 1, 1, 0, 0, 0, 66, True),  # SPO2,
        ("SPO2_quality", "SPO2_quality", "NONE", "na", 1., 1, 1, 0, 0, 0, 1002, True),  # SPO2_quality,
        ("BATT_ST_CHANNEL_CHAR", "BATT_ST_CHANNEL_CHAR", "", "na", 1., 1, 1, 0, 0, 0, 247, True),
        # BATT_ST_CHANNEL_CHAR,
        ("TEMP_ST_CHANNEL_CHAR", "TEMP_ST_CHANNEL_CHAR", "", "na", 1., 1, 1, 0, 0, 0, 246, True),
        # TEMP_ST_CHANNEL_CHAR,
        ("HRV_HF", "HRV_HF", "", "ms^2", 300., 300, 1. / 10., 0, 0, 8, 277, True),  # HRV_HF,
        ("HRV_LF_normalized", "HRV_LF_normalized", "", "", 300., 300, 1. / 100, 0, 0, 8, 273, True),
        # HRV_LF_normalized,
        ("HRV_LF", "HRV_LF", "", "ms^2", 300., 300, 1. / 10, 0, 0, 8, 276, True),  # HRV_LF,
        ("NN_over_RR", "NN_over_RR", "", "na", 300., 300, 1. / 100, 0, 0, 8, 274, True),  # NN_over_RR, #
        ("ANN", "ANN", "", "s", 300., 300, 1. / freq / 16., 0, 0, 8, 271, True),  # ANN,
        ("SDNN", "SDNN", "", "s", 300., 300, 1. / freq / 16., 0, 0, 8, 272, True),  # SDNN,
        ("HRV_triangular", "HRV_triangular", "", "na", 300., 300, 1. / 100., 0, 0, 5, 275, True),  # HRV_triangular,
        ("PPG", "PPG", "", "na", 4. / 256, 0, 1, 0, 0, 5, 64, True),  # PPG,
        ("QRS", "QRS_epoch", "", "na", 1., 0, 1., 0, 0, 0, 17, False),  # QRS
        ("RR_interval", "RR_interval_epoch", "RR_interval_quality", "s", 1., 0, 1 / freq, 0, 0, 8, 18, False),
        ("inspiration", "inspiration_epoch", "", "na", 1., 0, 1., 0, 0, 0, 34, False),  # inspiration
        ("expiration", "expiration_epoch", "", "", 1., 0, 1., 0, 0, 0, 35, False),  # expiration
        ("step", "step_epoch", "", "step", 1., 0, 1., 0, 0, 0, 52, False),  # step
        ("PTT", "PTT_epoch", "", "s", 1., 0, 1. / 256, 0, 0, 0, 97, False),  # PTT
        ("device_position", "device_position_epoch", "", "na", 1., 0, 1., 0, 0, 0, 269, False),  # device_position
        ("sleep_position", "sleep_position_epoch", "", "na", 1., 0, 1., 0, 0, 0, 270, False),  # sleep_position
        ("sleep_phase", "sleep_phase_epoch", "", "na", 1., 0, 1., 0, 0, 0, 280, False),  # sleep_phase
        ("NN_interval", "NN_interval_epoch", "", "s", 1., 0, 1 / freq, 0, 0, 8, 318, False),  # NN_interval
        ("RR_interval_realigned", "RR_interval_realigned_epoch", "", "s", 1., 0, 1 / freq, 0, 0, 8, 319, False),
        ("RR_interval_quality", "RR_interval_quality_epoch", "", "na", 1., 0, 1., 0, 0, 0, 1004, False),
    ]
    return specs

set_freq(256)


groups = [
    ('group_raw', 'outputDataRaw', raw, False),
    ('group', 'outputData', one_hz, True)
]


def _print_header(file_specs):
    return ",{} [{}] (/api/datatype/{}/)".format(file_specs['filename_in'], file_specs['unit'], file_specs['datatype'])


def _getspec(filepath):
    filename = os.path.splitext(os.path.split(filepath)[1])[0]
    info = load_info(os.path.dirname(filepath))
    freq = info['freq']
    try:
        out = next(a for a in files_specs_table(freq) if a[0] == filename)
    except StopIteration:
        raise AssertionError('Not a valid filename {}'.format(filepath))
    return {key: val for key, val in zip(keys_list, out)}


def _convert_file(dirname, dirname_out, settings, file_specs, extension):
    filepath = os.path.join(dirname, file_specs['filename_in'])
    data, data_quality = load_data(filepath, **settings)
    if settings['epoch'] and '_epoch' in file_specs['filename_out']:
        filepath_out = os.path.join(dirname_out, file_specs['filename_out'] + extension)
    elif (dirname != dirname_out or settings['start'] == 0) and not settings['epoch'] and '_epoch' in file_specs[
        'filename_out']:
        filepath_out = os.path.join(dirname_out, file_specs['filename_in'] + extension)
    elif '_epoch' in file_specs['filename_out']:
        pass
    else:
        filepath_out = os.path.join(dirname_out, file_specs['filename_out'] + extension)
    _save_file(data, data_quality, filepath_out, file_specs, **settings)


def load_data(filepath, start=0, add_quality=True, skip_raw=False, **settings):
    file_specs = _getspec(filepath)
    data = []
    data_quality = []
    # load_data
    if (skip_raw and file_specs['filename_in'] in raw) or \
            (add_quality and os.path.split(filepath)[1].endswith(
                '_quality')):  # if we save quality in heart_rate.csv we don` save heart_rate_quality.csv
        return data, data_quality
    if os.path.exists(filepath + '.wav'):
        data = load_wave(filepath + '.wav', start)
    elif os.path.exists(filepath + '.csv'):
        data = load_csv(filepath + '.csv', start)
    # load from file  _quality if necessary
    if add_quality and file_specs['quality_signal_name']:
        filepath = os.path.join(os.path.split(filepath)[0], file_specs['quality_signal_name'])
        if os.path.exists(filepath + '.wav'):
            data_quality = load_wave(filepath + '.wav', start)
        elif os.path.exists(filepath + '.csv'):
            data_quality = load_csv(filepath + '.csv', start)
    return data, data_quality


def _save_file(data, data_quality, filepath_out, file_specs, fmt=None, header_string=None, header=True, file_count=0,
               epoch=False, as_nan=False, delimiter_tab=False, **settings):
    delimiter = '\t' if delimiter_tab else ','
    if fmt is None:
        fmt = ['%.8f']
        fmt += ['%.{}f'.format(file_specs['round'])]
    if header:
        if header_string is None:
            header_string = 'time [s]'
            header_string += _print_header(file_specs) if header else ''
    header_string = header_string or ''
    if len(data_quality):
        data = _merge_array(data, data_quality, as_nan)
        filepath_q = os.path.join(os.path.split(filepath_out)[0], file_specs['quality_signal_name'])
        file_specs_quality = _getspec(filepath_q)
        header_string += _print_header(file_specs_quality) if header else ''
        fmt += ['%.{}f'.format(0)]
    if file_count == 0:
        mode = 'w'
    else:
        mode = 'a'
        header_string = ''
    if len(data):
        with open(filepath_out, mode) as f:
            f.write(header_string + '\n' if header else '')
            if as_nan:
                fmt = delimiter.join(fmt) + '\n'
                for line in data:
                    f.write(fmt % tuple(line))
            else:
                for line in data:
                    f.write(delimiter.join([fm % a if a is not None else '' for a, fm in zip(line, fmt)]) + '\n')


def load_csv(filepath, start=0):
    data = []
    file_specs = _getspec(filepath)
    with open(filepath, 'r') as csvfile:
        reader_csv = csv.reader(csvfile, delimiter=',', quotechar='|')
        for row in reader_csv:
            try:
                try:
                    int(row[1])  # ensure it is an int otherwise is was already converted, don't change second column
                    data.append([float(row[0]) + start, float(row[1]) * file_specs['gain'] + file_specs['offset']])
                except ValueError:
                    data.append([float(row[0]) + start, float(row[1])])
            except ValueError:
                pass
    return data


def load_wave(filepath, start=0):
    """
    Loads a WAV file according to the specified format. Time t=0 corresponds to
    the beginning of the record.
    """
    file_specs = _getspec(filepath)

    f = wave.open(filepath)
    if file_specs['signed']:
        yy = struct.unpack("h" * f.getnframes(), f.readframes(f.getnframes()))
    else:
        yy = struct.unpack("H" * f.getnframes(), f.readframes(f.getnframes()))
    f.close()
    return [(float(t) / f.getframerate() + file_specs['dt_offset'] + start,
             float(y) * file_specs['gain'] + file_specs['offset']) for t, y in enumerate(yy)]


def _convert_sub_directory(dirname, dirnameout, settings, freq):
    extension = '.txt' if settings.get('delimiter_tab', False) else '.csv'
    if settings['group'] or settings['group_raw']:
        for mode, filepath_out, channels, add_quality in groups:
            if settings[mode]:
                datas = None
                settings['add_quality'] = add_quality
                filepath_out = os.path.join(dirnameout, filepath_out + extension)
                fmt = ['%.8f']
                header = 'time [s]'
                for file_specs in files_specs_table(freq):
                    file_specs = {key: val for key, val in zip(keys_list, file_specs)}
                    if file_specs['filename_in'] in channels and not \
                            (not settings['add_quality'] and 'quality' in file_specs['filename_in']):
                        filepath = os.path.join(dirname, file_specs['filename_in'])
                        data, data_quality = load_data(filepath, **settings)
                        if len(data):
                            datas = _merge_array(datas, data, file_specs['dt'], **settings)
                            fmt += ['%.{}f'.format(file_specs['round'])]
                            header += _print_header(file_specs) if settings['header'] else ''
                        if len(data_quality):
                            datas = _merge_array(datas, data_quality, file_specs['dt'], **settings)
                            fmt += ['%.{}f'.format(0)]
                            filepath_q = os.path.join(dirname, file_specs['quality_signal_name'])
                            file_specs_quality = _getspec(filepath_q)
                            header += _print_header(file_specs_quality) if settings['header'] else ''
                _save_file(datas, [], filepath_out, file_specs, fmt=fmt, header_string=header, **settings)
    else:
        for file_specs in files_specs_table(freq):
            file_specs = {key: val for key, val in zip(keys_list, file_specs)}
            _convert_file(dirname, dirnameout, settings, file_specs, extension)


def _merge_array(in1, in2, dt=None, as_nan=False, **settings):
    if not in1:
        return in2
    in2 = _add_none(in2, dt, as_nan)
    return [l1 + l2[1:] for l1, l2 in zip(in1, in2)]


def _add_none(data, dt, as_nan):
    """allow to put 256, 128,64 Hz in the same file by filling with None"""
    count = {1 / 256: 0, 1 / 128: 1, 1 / 64: 3, 1 / 250: 0, 1 / 125: 1, 1 / 50: 4}
    if data:
        tmp = tuple([float('nan') for a in data[0]]) if as_nan else tuple([None for a in data[0]])
    for a in data:
        yield a
        for i in range(count.get(dt, 0)):
            yield tmp


def check_freq(info):
    root_start = info.get('start', info.get('start_timestamp', 0))
    if 1230768000 * 256 < root_start < 2524608000 * 256: # 2009, 2050
        return 256
    elif 1230768000 * 1000 < root_start < 2524608000 * 1000:  # 2009, 2050
        return 1000
    else:
        raise AssertionError("start_timestamp ={}. out of range. Aborted.".format(root_start))


def load_info(directory):
    with open(os.path.join(directory, "info.json"), "r") as f:
        info = json.load(f)
        info['freq'] = check_freq(info)
    return info


def convert_directory(directory, settings=None):
    settings = settings or DEFAULT_SETTINGS
    if not os.path.exists(os.path.join(directory, "info.json")):
        raise IOError("{} is missing. Please redownload. Aborted.".format(os.path.join(directory, "info.json")))
    check_version(directory)
    info = load_info(directory)
    freq = info['freq']
    root_start = info.get('start', info.get('start_timestamp', 0))

    dir_list = []
    for subdir in os.listdir(directory):
        subdir = os.path.join(directory, subdir)
        if os.path.isdir(subdir) and os.path.split(subdir)[1].startswith('record_'):
            with open(os.path.join(subdir, "info.json"), "r") as f:
                info_sub = json.load(f)
            dir_list.append([info_sub.get('start', info_sub.get('start_timestamp', 0)), subdir])
    if dir_list:
        dir_list = sorted(dir_list)
        for i, (start_subdir, subpath) in enumerate(dir_list):
            settings2 = copy.copy(settings)
            settings2['header'] = settings2['header'] and i == 0
            settings2['file_count'] = i
            settings2['start'] = start_subdir / freq if settings['epoch'] else (start_subdir - root_start) / freq
            _convert_sub_directory(subpath, directory, settings2, freq)
    else:
        settings['start'] = root_start / freq if settings['epoch'] else 0
        _convert_sub_directory(directory, directory, settings, freq)


def check_version(directory):
    with open(os.path.join(directory, "info.json"), "r") as f:
        info = json.load(f)
    file_version = [int(a) for a in info['decoder_version'].split('.')]
    if VERSION < file_version:
        raise AssertionError("Decoder version {}  too old. redownload decoder at: "
                             "'http://support.hexoskin.com/customer/portal/articles/1758581-how-do-i-convert-binary-"
                             "data-downloaded-from-the-dashboard-'\n".format(VERSION))
